pub const OTTR_TRIPLE: &str = "http://ns.ottr.xyz/0.4/Triple";
pub const OTTR_PREFIX: &str = "ottr";
pub const OTTR_PREFIX_IRI: &str = "http://ns.ottr.xyz/0.4/";
pub const RDF_PREFIX: &str = "rdf";
pub const RDF_PREFIX_IRI: &str = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
pub const RDFS_PREFIX: &str = "rdfs";
pub const RDFS_PREFIX_IRI: &str = "http://www.w3.org/2000/01/rdf-schema#";
pub const XSD_PREFIX: &str = "xsd";
pub const XSD_PREFIX_IRI: &str = "http://www.w3.org/2001/XMLSchema#";
pub const DEFAULT_PREFIX: &str = "urn:maplib_default:";
pub const OTTR_IRI: &str = "http://ns.ottr.xyz/0.4/IRI";
pub const OTTR_BLANK_NODE: &str = "http://ns.ottr.xyz/0.4/BlankNode";

pub const OWL_PREFIX: &str = "owl";
pub const OWL_PREFIX_IRI: &str = "http://www.w3.org/2002/07/owl#";

pub const SHACL_PREFIX: &str = "sh";

pub const SHACL_PREFIX_IRI: &str = "http://www.w3.org/ns/shacl#";

pub const XSD_ANY_URI: &str = "http://www.w3.org/2001/XMLSchema#anyURI";
